import base64


def image_bytes_to_base64_data_uri(img_bytes: bytes,
                                   format: str = "jpeg", encoding="utf-8"):
    base64_data = base64.b64encode(img_bytes).decode(encoding)
    return f"data:image/{format};base64,{base64_data}"


def image_to_base64_data_uri(file_path: str,
                             format: str = "jpeg", encoding="utf-8"):
    with open(file_path, "rb") as img_file:
        return image_bytes_to_base64_data_uri(
            img_file.read(), format=format, encoding=encoding)



from apipeline.pipeline.pipeline import Pipeline, FrameProcessor
from apipeline.pipeline.runner import PipelineRunner
from apipeline.pipeline.task import PipelineTask, FrameDirection, PipelineParams
from apipeline.frames.data_frames import Frame, TextFrame

from achatbot.processors.aggregators.llm_response import LLMAssistantResponseAggregator, LLMUserResponseAggregator
from achatbot.processors.aggregators.vision_image_frame import VisionImageFrameAggregator
from achatbot.processors.speech.tts.tts_processor import TTSProcessor
from achatbot.common.types import DailyParams
from achatbot.cmd.bots.base import DailyRoomBot
from achatbot.transports.daily import DailyTransport
from achatbot.types.frames.control_frames import UserImageRequestFrame
from .. import register_daily_room_bots


class UserImageRequester(FrameProcessor):
    def __init__(self, participant_id: str | None = None):
        super().__init__()
        self._participant_id = participant_id

    def set_participant_id(self, participant_id: str):
        self._participant_id = participant_id

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        if self._participant_id and isinstance(frame, TextFrame):
            await self.push_frame(UserImageRequestFrame(self._participant_id), FrameDirection.UPSTREAM)
        await self.push_frame(frame, direction)


@register_daily_room_bots.register
class DailyDescribeVisionBot(DailyRoomBot):
    async def arun(self):
        vad_analyzer = self.get_vad_analyzer()
        daily_params = DailyParams(
            audio_in_enabled=True,
            audio_out_enabled=True,
            vad_enabled=True,
            vad_analyzer=vad_analyzer,
            vad_audio_passthrough=True,
            transcription_enabled=False,
            camera_out_enabled=True,
            camera_out_is_live=True,
            camera_out_width=1280,
            camera_out_height=720
        )

        asr_processor = self.get_asr_processor()

        tts_processor: TTSProcessor = self.get_tts_processor()
        stream_info = tts_processor.get_stream_info()

        daily_params.audio_out_sample_rate = stream_info["sample_rate"]
        daily_params.audio_out_channels = stream_info["channels"]
        transport = DailyTransport(
            self.args.room_url, self.args.token, self.args.bot_name,
            daily_params,
        )

        llm_in_aggr = LLMUserResponseAggregator()
        image_requester = UserImageRequester()
        vision_aggregator = VisionImageFrameAggregator()
        llm_processor = self.get_llm_processor()
        llm_out_aggr = LLMAssistantResponseAggregator()

        @transport.event_handler("on_first_participant_joined")
        async def on_first_participant_joined(transport, participant):
            transport.capture_participant_video(participant["id"])
            transport.capture_participant_transcription(participant["id"])
            image_requester.set_participant_id(participant["id"])

        pipeline = Pipeline([
            transport.input_processor(),
            asr_processor,
            llm_in_aggr,
            image_requester,
            vision_aggregator,
            llm_processor,
            tts_processor,
            transport.output_processor(),
            llm_out_aggr,
        ])
        task = PipelineTask(pipeline, params=PipelineParams())
        await PipelineRunner().run(task)



from apipeline.pipeline.pipeline import Pipeline
from apipeline.pipeline.runner import PipelineRunner
from apipeline.pipeline.task import PipelineTask

from achatbot.common.types import DailyParams
from achatbot.cmd.bots.base import DailyRoomBot
from achatbot.transports.daily import DailyTransport
from .. import register_daily_room_bots


@register_daily_room_bots.register
class DailyEchoVisionBot(DailyRoomBot):
    async def arun(self):
        transport = DailyTransport(
            self.args.room_url, self.args.token, self.args.bot_name,
            DailyParams(
                audio_in_enabled=True,
                audio_out_enabled=True,
                camera_out_enabled=True,
                camera_out_is_live=True,
                camera_out_width=1280,
                camera_out_height=720
            )
        )

        @transport.event_handler("on_first_participant_joined")
        async def on_first_participant_joined(transport, participant):
            transport.capture_participant_video(participant["id"])

        pipeline = Pipeline([
            transport.input_processor(),
            transport.output_processor(),
        ])
        task = PipelineTask(pipeline)
        await PipelineRunner().run(task)



import logging
import base64
from io import BytesIO
from typing import AsyncGenerator
import uuid

from PIL import Image
from apipeline.frames.sys_frames import ErrorFrame
from apipeline.frames.data_frames import Frame, TextFrame

from achatbot.common.utils.img_utils import image_bytes_to_base64_data_uri
from achatbot.common.session import Session
from achatbot.common.types import SessionCtx
from achatbot.processors.vision.base import VisionProcessorBase
from achatbot.common.factory import EngineClass
from achatbot.common.interface import ILlm
from achatbot.types.frames.data_frames import VisionImageRawFrame


class LLamaCPPVisionProcessor(VisionProcessorBase):

    def __init__(
        self,
        llm: ILlm | EngineClass | None = None,
        session: Session | None = None,
    ):
        super().__init__()
        self._llm = llm
        self._session = session
        if self._session is None:
            self._session = Session(**SessionCtx(uuid.uuid4()).__dict__)

    def set_llm(self, llm: ILlm):
        self._llm = llm

    async def run_vision(self, frame: VisionImageRawFrame) -> AsyncGenerator[Frame, None]:
        """
        !TODO: image frame: PIL.Image, URL(str), base64 img(str) @weedge
        """
        if not self._llm:
            logging.error(f"{self} error: llm not available")
            yield ErrorFrame("llm not available")
            return

        logging.debug(f"Analyzing image: {frame}")

        image = Image.frombytes(frame.mode, frame.size, frame.image)
        buffered = BytesIO()
        image.save(buffered, format=frame.format)
        img_base64_str = image_bytes_to_base64_data_uri(buffered.getvalue(), frame.format.lower())

        self._session.ctx.state["prompt"] = [
            {"type": "text", "text": frame.text},
            {"type": "image_url", "image_url": {"url": img_base64_str}},
        ]

        iter = self._llm.chat_completion(self._session)
        for item in iter:
            yield TextFrame(text=item)



from abc import abstractmethod
import logging

from apipeline.processors.frame_processor import FrameDirection

from achatbot.processors.aggregators.openai_llm_context import OpenAILLMContext
from achatbot.processors.ai_processor import AIProcessor
from achatbot.types.frames.control_frames import UserImageRequestFrame


class LLMProcessor(AIProcessor):
    """This class is a no-op but serves as a base class for LLM processors."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._callbacks = {}
        self._start_callbacks = {}

    @abstractmethod
    async def set_model(self, model: str):
        pass

    @abstractmethod
    async def set_llm_args(self, **args):
        pass

    # !TODO: use callback function type @weedge
    def register_function(self, function_name: str | None, callback, start_callback=None):

        # Registering a function with the function_name set to None will run that callback
        # for all functions
        self._callbacks[function_name] = callback
        # QUESTION FOR CB: maybe this isn't needed anymore?
        if start_callback:
            self._start_callbacks[function_name] = start_callback

    def unregister_function(self, function_name: str | None):
        del self._callbacks[function_name]
        if self._start_callbacks[function_name]:
            del self._start_callbacks[function_name]

    def has_function(self, function_name: str):
        if None in self._callbacks.keys():
            return True
        return function_name in self._callbacks.keys()

    async def call_function(
            self,
            *,
            context: OpenAILLMContext,
            tool_call_id: str,
            function_name: str,
            arguments: str) -> None:
        f = None
        if function_name in self._callbacks.keys():
            f = self._callbacks[function_name]
        elif None in self._callbacks.keys():
            f = self._callbacks[None]
        else:
            return None
        await context.call_function(
            f,
            function_name=function_name,
            tool_call_id=tool_call_id,
            arguments=arguments,
            llm=self)

    # QUESTION FOR CB: maybe this isn't needed anymore?
    async def call_start_function(self, context: OpenAILLMContext, function_name: str):
        if function_name in self._start_callbacks.keys():
            await self._start_callbacks[function_name](function_name, self, context)
        elif None in self._start_callbacks.keys():
            return await self._start_callbacks[None](function_name, self, context)

    async def request_image_frame(self, user_id: str, *, text_content: str | None = None):
        await self.push_frame(
            UserImageRequestFrame(user_id=user_id, context=text_content),
            FrameDirection.UPSTREAM)
